@@ -24,9 +24,14 @@
 #include <config.h>
 #include <atk/atkcomponent.h>
 #include <atk/atktext.h>
+#include <glib/gi18n.h>
 
 #include "gtkhtml.h"
 #include "htmlengine.h"
+#include "htmlengine-edit.h"
+#include "htmlengine-edit-cursor.h"
+#include "htmlengine-edit-movement.h"
+#include "htmlengine-edit-cut-and-paste.h"
 #include "htmlinterval.h"
 #include "htmlselection.h"
 #include "htmltext.h"
@@ -44,6 +49,8 @@ static void atk_text_interface_init      (AtkTextIface *iface);
 static void html_a11y_text_get_extents   (AtkComponent *component,
 					  gint *x, gint *y, gint *width, gint *height, AtkCoordType coord_type);
 static void html_a11y_text_get_size      (AtkComponent *component, gint *width, gint *height);
+static gboolean html_a11y_text_grab_focus (AtkComponent *comp);
+
 static gchar * html_a11y_text_get_text (AtkText *text, gint start_offset, gint end_offset);
 static gchar * html_a11y_text_get_text_after_offset (AtkText *text, gint offset, AtkTextBoundary boundary_type,
 						     gint *start_offset, gint *end_offset);
@@ -80,15 +87,48 @@ static void	html_a11y_text_delete_text	(AtkEditableText      *text,
 						 gint                 end_pos);
 static void	html_a11y_text_paste_text	(AtkEditableText      *text,
 						 gint                 position);
-static void	html_a11y_text_paste_received	(GtkClipboard *clipboard,
-						 const gchar  *text,
-						 gpointer     data);
 
 static AtkStateSet* html_a11y_text_ref_state_set	(AtkObject	*accessible);
 
-
 static AtkObjectClass *parent_class = NULL;
 
+static gint
+get_n_actions (AtkAction *action)
+{
+        return 1;
+}
+
+static G_CONST_RETURN gchar*
+action_get_name (AtkAction *action, gint      i)
+{
+        if (i == 0)
+                return _("grab focus");
+
+        return NULL;
+}
+
+static gboolean
+do_action (AtkAction * action, gint i)
+{
+        switch (i) {
+        case 0:
+		return html_a11y_text_grab_focus (ATK_COMPONENT (action));
+        default:
+                return FALSE;
+        }
+}
+
+static void
+atk_action_interface_init (AtkActionIface *iface)
+{
+        g_return_if_fail (iface != NULL);
+
+        iface->do_action = do_action;
+        iface->get_n_actions = get_n_actions;
+        iface->get_name = action_get_name;
+}
+
+
 GType
 html_a11y_text_get_type (void)
 {
@@ -127,10 +167,18 @@ html_a11y_text_get_type (void)
 			NULL
 		};
 
+		static const GInterfaceInfo atk_action_info = {
+			(GInterfaceInitFunc) atk_action_interface_init,
+			(GInterfaceFinalizeFunc) NULL,
+			NULL
+		};
+
+
 		type = g_type_register_static (G_TYPE_HTML_A11Y, "HTMLA11YText", &tinfo, 0);
 		g_type_add_interface_static (type, ATK_TYPE_COMPONENT, &atk_component_info);
 		g_type_add_interface_static (type, ATK_TYPE_TEXT, &atk_text_info);
 		g_type_add_interface_static (type, ATK_TYPE_EDITABLE_TEXT, &atk_editable_text_info);
+		g_type_add_interface_static (type, ATK_TYPE_ACTION, &atk_action_info);
 	}
 
 	return type;
@@ -143,6 +191,7 @@ atk_component_interface_init (AtkComponentIface *iface)
 
 	iface->get_extents = html_a11y_text_get_extents;
 	iface->get_size = html_a11y_text_get_size;
+	iface->grab_focus = html_a11y_text_grab_focus;
 }
 
 static void
@@ -247,7 +296,8 @@ html_a11y_text_ref_state_set (AtkObject *accessible)
 		atk_state_set_add_state (state_set, ATK_STATE_EDITABLE);
 
 	atk_state_set_add_state (state_set, ATK_STATE_MULTI_LINE);
-	atk_state_set_add_state (state_set, ATK_STATE_MULTI_LINE);
+	atk_state_set_add_state (state_set, ATK_STATE_SENSITIVE);
+	atk_state_set_add_state (state_set, ATK_STATE_FOCUSABLE);
 
 	return state_set;
 }
@@ -296,6 +346,21 @@ html_a11y_text_get_size (AtkComponent *component, gint *width, gint *height)
 	get_size (obj, width, height);
 }
 
+
+static gboolean
+html_a11y_text_grab_focus (AtkComponent *comp)
+{
+	GtkHTML *html;
+
+	html = GTK_HTML_A11Y_GTKHTML (html_a11y_get_gtkhtml_parent (HTML_A11Y (comp)));
+	g_return_val_if_fail (html && html->engine && html_engine_get_editable (html->engine), FALSE);
+
+	html_engine_jump_to_object (html->engine, HTML_A11Y_HTML (comp), HTML_TEXT (HTML_A11Y_HTML (comp))->text_len);
+	g_signal_emit_by_name (html, "grab_focus");
+
+        return TRUE;
+}
+
 /*
  * AtkText interface
  */
@@ -306,6 +371,8 @@ html_a11y_text_get_text (AtkText *text, gint start_offset, gint end_offset)
 	HTMLText *to = HTML_TEXT (HTML_A11Y_HTML (text));
 	gchar *str;
 
+	g_return_val_if_fail (to, NULL);
+
 	/* printf ("%d - %d\n", start_offset, end_offset); */
 	if (end_offset == -1)
 		end_offset = to->text_len;
@@ -365,22 +432,95 @@ html_a11y_text_get_text_after_offset (AtkText *text, gint offset, AtkTextBoundar
 				      gint *start_offset, gint *end_offset)
 {
 	HTMLText *to = HTML_TEXT (HTML_A11Y_HTML (text));
+	HTMLTextSlave *start_slave, *end_slave;
 
-	gail_text_util_text_setup (HTML_A11Y_TEXT (text)->util, to->text);
-	return gail_text_util_get_text (HTML_A11Y_TEXT (text)->util, NULL, GAIL_AFTER_OFFSET, boundary_type, offset, 
+	g_return_val_if_fail (to, NULL);
+	g_return_val_if_fail (start_offset && end_offset, NULL);
+
+	switch (boundary_type) {
+	case ATK_TEXT_BOUNDARY_LINE_START:
+		end_slave = html_text_get_slave_at_offset (HTML_OBJECT (to), offset);
+		g_return_val_if_fail (end_slave, NULL);
+		start_slave = (HTMLTextSlave *) HTML_OBJECT (end_slave)->next;
+
+		if (start_slave && HTML_IS_TEXT_SLAVE (start_slave)) {
+			*start_offset = start_slave->posStart;
+			end_slave = (HTMLTextSlave *) HTML_OBJECT (start_slave)->next;
+			if (end_slave && HTML_IS_TEXT_SLAVE (end_slave)) 
+				*end_offset = end_slave->posStart;
+			else 
+				*end_offset = start_slave->posStart + start_slave->posLen;
+
+		} else { /* we are on the last line. */
+			*start_offset = *end_offset = html_a11y_text_get_character_count (text);
+		}
+
+		return html_a11y_text_get_text (text, *start_offset, *end_offset);
+
+	case ATK_TEXT_BOUNDARY_LINE_END:
+		start_slave = html_text_get_slave_at_offset (HTML_OBJECT (to), offset);
+		g_return_val_if_fail (start_slave, NULL);
+
+		*start_offset = start_slave->posStart + start_slave->posLen;
+		end_slave = (HTMLTextSlave *) HTML_OBJECT (start_slave)->next;
+		if (end_slave && HTML_IS_TEXT_SLAVE (end_slave))
+			*end_offset = end_slave->posStart + end_slave->posLen;
+		else
+			*end_offset = *start_offset;
+		
+		return html_a11y_text_get_text (text, *start_offset, *end_offset);
+
+	default:
+		gail_text_util_text_setup (HTML_A11Y_TEXT (text)->util, to->text);
+		return gail_text_util_get_text (HTML_A11Y_TEXT (text)->util, NULL, GAIL_AFTER_OFFSET, boundary_type, offset, 
 					start_offset, end_offset);
+	}
 }
 
 static gchar *
 html_a11y_text_get_text_at_offset (AtkText *text, gint offset, AtkTextBoundary boundary_type,
 				   gint *start_offset, gint *end_offset)
 {
-	gchar * ret;
 	HTMLText *to = HTML_TEXT (HTML_A11Y_HTML (text));
+	HTMLTextSlave *start_slave, *end_slave;
 
-	gail_text_util_text_setup (HTML_A11Y_TEXT (text)->util, to->text);
-	return gail_text_util_get_text (HTML_A11Y_TEXT (text)->util, NULL, GAIL_AT_OFFSET, boundary_type, offset, 
+	g_return_val_if_fail (to, NULL);
+	g_return_val_if_fail (start_offset && end_offset, NULL);
+
+	switch (boundary_type) {
+	case ATK_TEXT_BOUNDARY_LINE_START:
+		start_slave = html_text_get_slave_at_offset (HTML_OBJECT (to), offset);
+		g_return_val_if_fail (start_slave, NULL);
+		end_slave = (HTMLTextSlave *) HTML_OBJECT (start_slave)->next;
+
+		if (end_slave && HTML_IS_TEXT_SLAVE (end_slave)) {
+			*end_offset = end_slave->posStart;
+		} else {
+			*end_offset = start_slave->posStart + start_slave->posLen;
+		}
+		*start_offset = start_slave->posStart;
+
+		return html_a11y_text_get_text (text, *start_offset, *end_offset);
+
+	case ATK_TEXT_BOUNDARY_LINE_END:
+		end_slave = html_text_get_slave_at_offset (HTML_OBJECT (to), offset);
+		g_return_val_if_fail (end_slave, NULL);
+		start_slave = (HTMLTextSlave *) HTML_OBJECT (end_slave)->prev;
+
+		if (start_slave && HTML_IS_TEXT_SLAVE (start_slave)) {
+			*start_offset = start_slave->posStart + start_slave->posLen;
+		} else {
+			*start_offset = end_slave->posStart;
+		}
+		*end_offset = end_slave->posStart + end_slave->posLen;
+
+		return html_a11y_text_get_text (text, *start_offset, *end_offset);
+
+	default:
+		gail_text_util_text_setup (HTML_A11Y_TEXT (text)->util, to->text);
+		return gail_text_util_get_text (HTML_A11Y_TEXT (text)->util, NULL, GAIL_AT_OFFSET, boundary_type, offset, 
 					start_offset, end_offset);
+	}
 	
 }
 
@@ -389,7 +529,7 @@ html_a11y_text_get_character_at_offset (AtkText *text, gint offset)
 {
 	HTMLText *to = HTML_TEXT (HTML_A11Y_HTML (text));
 
-	g_return_val_if_fail (offset <= to->text_len, 0);
+	g_return_val_if_fail (to && offset <= to->text_len, 0);
 
 	return html_text_get_char (to, offset);
 }
@@ -399,25 +539,67 @@ html_a11y_text_get_text_before_offset (AtkText *text, gint offset, AtkTextBounda
 				       gint *start_offset, gint *end_offset)
 {
 	HTMLText *to = HTML_TEXT (HTML_A11Y_HTML (text));
+	HTMLTextSlave *start_slave, *end_slave;
 
 	g_return_val_if_fail (to, NULL);
+	g_return_val_if_fail (start_offset && end_offset, NULL);
 
-	gail_text_util_text_setup (HTML_A11Y_TEXT (text)->util, to->text);
-	
-	return gail_text_util_get_text (HTML_A11Y_TEXT (text)->util, NULL, GAIL_BEFORE_OFFSET, boundary_type, offset, 
+	switch (boundary_type) {
+	case ATK_TEXT_BOUNDARY_LINE_START:
+		end_slave = html_text_get_slave_at_offset (HTML_OBJECT (to), offset);
+		g_return_val_if_fail (end_slave, NULL);
+		start_slave = (HTMLTextSlave *) HTML_OBJECT (end_slave)->prev;
+
+		*end_offset = end_slave->posStart;
+		if (start_slave && HTML_IS_TEXT_SLAVE (start_slave)) {
+			*start_offset = start_slave->posStart;
+		} else 
+			*start_offset = *end_offset;
+
+		return html_a11y_text_get_text (text, *start_offset, *end_offset);
+
+	case ATK_TEXT_BOUNDARY_LINE_END:
+		start_slave = html_text_get_slave_at_offset (HTML_OBJECT (to), offset);
+		g_return_val_if_fail (start_slave, NULL);
+		end_slave = (HTMLTextSlave *) HTML_OBJECT (start_slave)->prev;
+
+		if (end_slave && HTML_IS_TEXT_SLAVE (end_slave)) {
+			*end_offset = end_slave->posStart + end_slave->posLen;
+			start_slave = (HTMLTextSlave *) HTML_OBJECT (end_slave)->prev;
+			if (start_slave && HTML_IS_TEXT_SLAVE (start_slave))
+				*start_offset = start_slave->posStart + start_slave->posLen;
+			else 
+				*start_offset = end_slave->posStart;
+
+		} else {
+			*start_offset = *end_offset = 0;	/* on the first line */
+		}
+
+		return html_a11y_text_get_text (text, *start_offset, *end_offset);
+
+	default:
+		gail_text_util_text_setup (HTML_A11Y_TEXT (text)->util, to->text);
+		return gail_text_util_get_text (HTML_A11Y_TEXT (text)->util, NULL, GAIL_BEFORE_OFFSET, boundary_type, offset, 
 					start_offset, end_offset);
+	}
 }
 
 static gint
 html_a11y_text_get_character_count (AtkText *text)
 {
-	return HTML_TEXT (HTML_A11Y_HTML (text))->text_len;
+	HTMLText *to = HTML_TEXT (HTML_A11Y_HTML (text));
+
+	g_return_val_if_fail (to, 0);
+	return to->text_len;
 }
 
 static gint
 html_a11y_text_get_n_selections (AtkText *text)
 {
-	return HTML_A11Y_HTML (text)->selected ? 1 : 0;
+	HTMLObject *to = HTML_A11Y_HTML (text);
+
+	g_return_val_if_fail (to, 0);
+	return to->selected ? 1 : 0;
 }
 
 static gchar *
@@ -425,7 +607,7 @@ html_a11y_text_get_selection (AtkText *text, gint selection_num, gint *start_off
 {
 	HTMLText *to = HTML_TEXT (HTML_A11Y_HTML (text));
 
-	if (!HTML_OBJECT (to)->selected || selection_num > 0)
+	if (!to || !HTML_OBJECT (to)->selected || selection_num > 0)
 		return NULL;
 
 	*start_offset = to->select_start;
@@ -545,7 +727,6 @@ html_a11y_text_insert_text (AtkEditableText *text,
 {
 	GtkHTML * html;
 	HTMLText *t;
-	gint index;
 
 	/* fprintf(stderr, "atk insert text called \n"); */
 
@@ -565,9 +746,7 @@ html_a11y_text_copy_text	(AtkEditableText *text,
 				 gint            end_pos)
 {
 	GtkHTML * html;
-	gchar * str;
 	HTMLText *t;
-	gint start_index, end_index;
 
 	/* fprintf(stderr, "atk copy text called \n"); */
         html = GTK_HTML_A11Y_GTKHTML(html_a11y_get_gtkhtml_parent(HTML_A11Y(text)));
@@ -592,8 +771,6 @@ html_a11y_text_cut_text (AtkEditableText *text,
 {
 	GtkHTML * html;
 	HTMLText *t;
-	gint start_index, end_index;
-	gchar * str;
 
 	/* fprintf(stderr, "atk cut text called.\n"); */
         html = GTK_HTML_A11Y_GTKHTML(html_a11y_get_gtkhtml_parent(HTML_A11Y(text)));
@@ -619,8 +796,6 @@ html_a11y_text_delete_text	(AtkEditableText *text,
 {
 	GtkHTML * html;
 	HTMLText *t;
-	gint start_index, end_index;
-	gchar * str;
 
 	/* fprintf(stderr, "atk delete text called.\n"); */
         html = GTK_HTML_A11Y_GTKHTML(html_a11y_get_gtkhtml_parent(HTML_A11Y(text)));
