@@ -27,7 +27,7 @@
 #include <config.h>
 #endif
 
-#include <libgnome/gnome-i18n.h>
+#include <glib/gi18n.h>
 #include <string.h>
 #include <gnome.h>
 #include <bonobo.h>
