@@ -1768,6 +1768,47 @@ append_selection_string (HTMLObject *self,
 	}
 }
 
+/*
+ * Calculate given UTF-8 string's "cell width", determined by
+ * g_unichar_iswide() on each character.  (So it doesn't work on
+ * zero-width or combined characters.)
+ */
+static gint
+utf8_width (const char *str, gint len)
+{
+	gunichar c;
+	gint width = 0;
+
+	while (len--) {
+		c = g_utf8_get_char (str);
+		width += g_unichar_iswide (c) ? 2 : 1;
+		str = g_utf8_next_char (str);
+	}
+	return width;
+}
+
+/*
+ * Returns the length of string starting on the given string which is
+ * not exceeding the given width.
+ */
+static gint
+utf8_length_in_width (const char *str, gint len, gint width)
+{
+	gunichar c;
+	gint l = 0;
+
+	while (len--) {
+		c = g_utf8_get_char (str);
+		width -= g_unichar_iswide (c) ? 2 : 1;
+		if (width < 0)
+			break;
+		str = g_utf8_next_char (str);
+		l++;
+	}
+
+	return l;
+}
+
 static gboolean
 save_plain (HTMLObject *self,
 	    HTMLEngineSaveState *state,
@@ -1779,18 +1820,18 @@ save_plain (HTMLObject *self,
 	gint pad;
 	gint align_pad;
 	gboolean firstline = TRUE;
-	gint max_len;
+	gint max_width;
 
 	flow = HTML_CLUEFLOW (self);
 
 	pad = plain_padding (flow, NULL, FALSE);
 	buffer_state = html_engine_save_buffer_new (state->engine, 
 						    state->inline_frames);
-	max_len = MAX (requested_width - pad, 0);
+	max_width = MAX (requested_width - pad, 0);
 	/* buffer the paragraph's content into the save buffer */
 	if (HTML_OBJECT_CLASS (&html_clue_class)->save_plain (self, 
 							      buffer_state, 
-							      max_len)) {
+							      max_width)) {
 		guchar *s;
 		int offset;
 		
@@ -1811,7 +1852,7 @@ save_plain (HTMLObject *self,
 			PangoContext *pc = state->engine->painter->pango_context;
 			PangoLogAttr *lattrs;
 			PangoItem **items;
-			gint len, skip;
+			gint len, width, skip;
 
 			items_list = pango_itemize (pc, s, 0, bytes, attrs, NULL);
 			lattrs = g_new (PangoLogAttr, slen + 1);
@@ -1854,15 +1895,18 @@ save_plain (HTMLObject *self,
 			while (*s) {
 				len = strcspn (s, "\n");
 				len = g_utf8_strlen (s, len);
+				width = utf8_width (s, len);
 				skip = 0;
 			
 				if ((flow->style != HTML_CLUEFLOW_STYLE_PRE) 
 				    && !HTML_IS_TABLE (HTML_CLUE (flow)->head)) {
-					if (len > max_len) {
+					if (width > max_width) {
 						gboolean look_backward = TRUE;
+						gint wmax;
 						gint wi, wl;
 
-						wl = clen + max_len;
+						wmax = clen + utf8_length_in_width (s, len, max_width);
+						wl = wmax;
 
 						if (lattrs [wl].is_white) {
 
@@ -1872,7 +1916,7 @@ save_plain (HTMLObject *self,
 							if (wl < slen && html_text_is_line_break (lattrs [wl]))
 								look_backward = FALSE;
 							else
-								wl = clen + max_len;
+								wl = wmax;
 						}
 
 						if (look_backward) {
@@ -1884,10 +1928,11 @@ save_plain (HTMLObject *self,
 						}
 
 						if (wl > clen && wl < slen && html_text_is_line_break (lattrs [wl])) {
-							wi = MIN (wl, clen + max_len);
-							while (wi > 0 && lattrs [wi - 1].is_white)
+							wi = MIN (wl, wmax);
+							while (wi > clen && lattrs [wi - 1].is_white)
 								wi --;
 							len = wi - clen;
+							width = utf8_width (s, len);
 							skip = wl - wi;
 						}
 					}
@@ -1900,10 +1945,10 @@ save_plain (HTMLObject *self,
 
 				switch (html_clueflow_get_halignment (flow)) {
 				case HTML_HALIGN_RIGHT:
-					align_pad = max_len - len;
+					align_pad = max_width - width;
 					break;
 				case HTML_HALIGN_CENTER:
-					align_pad = (max_len - len) / 2;
+					align_pad = (max_width - width) / 2;
 					break;
 				default:
 					align_pad = 0;
@@ -2258,6 +2303,39 @@ html_clueflow_type_init (void)
 	html_clueflow_class_init (&html_clueflow_class, HTML_TYPE_CLUEFLOW, sizeof (HTMLClueFlow));
 }
 
+static HTMLDirection
+html_clueflow_real_get_direction (HTMLObject *o)
+{
+	HTMLDirection dir = HTML_CLUEFLOW (o)->dir;
+
+	if (dir == HTML_DIRECTION_DERIVED && o->parent) {
+		dir = html_object_get_direction (o->parent);
+	}
+
+	if (dir == HTML_DIRECTION_DERIVED) {
+		HTMLObject *child;
+		for (child = HTML_CLUE (o)->head; child; child = child->next) {
+			if (HTML_IS_TEXT (child)) {
+				PangoDirection pdir = html_text_get_pango_direction (HTML_TEXT (child));
+				switch (pdir) {
+				case PANGO_DIRECTION_RTL:
+					dir = HTML_DIRECTION_RTL;
+					break;
+				case PANGO_DIRECTION_LTR:
+					dir = HTML_DIRECTION_LTR;
+					break;
+				default:
+					break;
+				}
+
+				break;
+			}
+		}
+	}
+
+	return dir;
+}
+
 void
 html_clueflow_class_init (HTMLClueFlowClass *klass,
 			  HTMLType type,
@@ -2293,6 +2371,7 @@ html_clueflow_class_init (HTMLClueFlowClass *klass,
 	object_class->get_recursive_length = get_recursive_length;
 	object_class->get_clear = get_clear;
 	object_class->set_painter = set_painter;
+	object_class->get_direction = html_clueflow_real_get_direction;
 
 	klass->get_default_font_style = get_default_font_style;
 
@@ -2316,6 +2395,7 @@ html_clueflow_init (HTMLClueFlow *clueflow, HTMLClueFlowClass *klass,
 
 	clue->valign = HTML_VALIGN_BOTTOM;
 	clue->halign = HTML_HALIGN_NONE;
+	clueflow->dir = HTML_DIRECTION_DERIVED;
 
 	clueflow->style = style;
 	clueflow->levels = levels; 
@@ -2553,13 +2633,30 @@ html_clueflow_get_halignment (HTMLClueFlow *flow)
 	g_return_val_if_fail (flow != NULL, HTML_HALIGN_NONE);
 
 	if (HTML_CLUE (flow)->halign == HTML_HALIGN_NONE) {
+		HTMLHAlignType halign;
+
 		if (HTML_OBJECT (flow)->parent && HTML_IS_TABLE_CELL (HTML_OBJECT (flow)->parent))
-			return HTML_CLUE (HTML_OBJECT (flow)->parent)->halign == HTML_HALIGN_NONE
-				? HTML_TABLE_CELL (HTML_OBJECT (flow)->parent)->heading ? HTML_HALIGN_CENTER : HTML_HALIGN_LEFT
+			halign = HTML_CLUE (HTML_OBJECT (flow)->parent)->halign == HTML_HALIGN_NONE
+				? HTML_TABLE_CELL (HTML_OBJECT (flow)->parent)->heading ? HTML_HALIGN_CENTER : HTML_HALIGN_NONE
 				: HTML_CLUE (HTML_OBJECT (flow)->parent)->halign;
 		else
-			return HTML_CLUE (HTML_OBJECT (flow)->parent)->halign == HTML_HALIGN_NONE
-				? HTML_HALIGN_LEFT : HTML_CLUE (HTML_OBJECT (flow)->parent)->halign;
+			halign = HTML_CLUE (HTML_OBJECT (flow)->parent)->halign;
+
+		if (halign == HTML_HALIGN_NONE) {
+			switch (html_object_get_direction (HTML_OBJECT (flow))) {
+			case HTML_DIRECTION_LTR:
+				halign = HTML_HALIGN_LEFT;
+				break;
+			case HTML_DIRECTION_RTL:
+				halign = HTML_HALIGN_RIGHT;
+				break;
+			default:
+				break;
+			}
+		}
+
+		return halign;
+
 	} else
 		return HTML_CLUE (flow)->halign;
 }
